#!/usr/bin/env python

KEY_SPACING_X=1
KEY_SPACING_Y=1

import cairo
import math
import struct

import keyclasses
import keyboards

KEYPLACES=None
CURRENT_KEYBOARD_SET=None

def put_text(ctx, text, tx, ty):
    ctx.set_source_rgba(*CURRENT_KEYBOARD_SET['TEXT_COLORBG'])
    ctx.move_to(tx-1,ty-1)
    ctx.show_text(text)
    ctx.move_to(tx+1,ty+1)
    ctx.show_text(text)
    ctx.move_to(tx-1,ty+1)
    ctx.show_text(text)
    ctx.move_to(tx+1,ty-1)
    ctx.show_text(text)
    ctx.move_to(tx,ty)
    ctx.set_source_rgba(*CURRENT_KEYBOARD_SET['TEXT_COLOR'])
    ctx.show_text(text)
    

def draw_button(ctx, text, size, x, y, w, h, text2=None):
    ctx.set_source_rgba(*CURRENT_KEYBOARD_SET['BORDER_COLOR2'])
    ctx.set_line_width(1)
    ctx.move_to(x+1, y+1)
    ctx.rel_line_to(w-2, 0)
    ctx.rel_line_to(0, h-2)
    ctx.rel_line_to(-w+2, 0)
    ctx.close_path()
    ctx.fill()
    ctx.set_source_rgba(*CURRENT_KEYBOARD_SET['BORDER_COLOR'])
    ctx.set_line_width(1)
    ctx.move_to(x, y)
    ctx.rel_line_to(w, 0)
    ctx.rel_line_to(0, h)
    ctx.rel_line_to(-w, 0)
    ctx.close_path()
    ctx.stroke()
    ctx.select_font_face("Sans")
    ctx.set_font_size(size)
    fascent, fdescent, fheight, fxadvance, fyadvance = ctx.font_extents()
    xbearing, ybearing, width, height, xadvance, yadvance = ctx.text_extents(text)
    tx=x-xbearing-width/2+w/2
    ty=y-fdescent+fheight/2+h/2
    put_text(ctx, text, tx, ty)
    if text2!=None:
        ctx.set_font_size(int(size/3*2))
        fascent, fdescent, fheight, fxadvance, fyadvance = ctx.font_extents()
        xbearing, ybearing, width, height, xadvance, yadvance = ctx.text_extents(text2)
        tx=x-xbearing-width/2+w*3/4
        ty=y-fdescent+fheight/2+h/4
        put_text(ctx, text2, tx, ty)

keyclasses.draw_button=draw_button

def place_keyblock(block, dimensions):
    startx,starty,width,height=dimensions
    lines=float(len(block))
    keyh=(height-(lines-1)*KEY_SPACING_Y)/lines
    y=starty
    for line in block:
        x=startx
        sumsize=float(sum([k.get_size() for k in line]))
        keyw=(width-(len(line)-1)*KEY_SPACING_X)/sumsize
        for k in line:
            w=keyw*k.get_size()
            k.set_placement(x,y,w,keyh)
            chars=k.get_chars()
            if chars!=None:
                if isinstance(chars, str):
                    chars=keyclasses.SPECIAL_KEYCODES[chars]
                KEYPLACES.append([int(round(x)),int(round(y)),int(round(w)),int(round(keyh)),
                                  chars])
            x+=w+KEY_SPACING_X
        y+=keyh+KEY_SPACING_Y

def draw_keyblock(ctx, block):
    # block must be already placed
    for line in block:
        for key in line:
            x,y,w,h=key.get_placement()
            key.render(ctx,x,y,w,h)


def make_one_keyboard(keyboard_set):
    global CURRENT_KEYBOARD_SET
    global KEYPLACES
    CURRENT_KEYBOARD_SET = keyboard_set
    KEYPLACES = []
    surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, CURRENT_KEYBOARD_SET['width'], CURRENT_KEYBOARD_SET['height'])
    ctx = cairo.Context(surface)
    for (block,dim) in CURRENT_KEYBOARD_SET['blocks']:
        place_keyblock(block, dim)
        draw_keyblock(ctx, block)
    surface.write_to_png("keyboard-"+CURRENT_KEYBOARD_SET['name']+".png")
    kpf=open('keyboard-'+CURRENT_KEYBOARD_SET['name']+'.places','w')
    keycodes_table=[]
    for keyplace in KEYPLACES:
        assert len(keyplace)==5
        assert isinstance(keyplace[0], int)
        assert isinstance(keyplace[1], int)
        assert isinstance(keyplace[2], int)
        assert isinstance(keyplace[3], int)
        assert isinstance(keyplace[4], int) or isinstance(keyplace[4], list)
        if isinstance(keyplace[4], list):
            assert len(keyplace[4])==4
            assert isinstance(keyplace[4][0], list)
            assert isinstance(keyplace[4][1], list)
            assert isinstance(keyplace[4][2], list)
            assert isinstance(keyplace[4][3], list)
        # specification of binary format for key positions:
        # file consists of records with fixed length, each record has this format (all values 16bit LE unsigned):
        # addr 0: uint16 x position
        # addr 2: uint16 y position
        # addr 4: uint16 width
        # addr 6: uint16 height
        # addr 8: uint16 keynr - key number,
        #                 this number is index into keycodes table
        #                 or special key directly for values >= 60000
        if isinstance(keyplace[4], list):
            keynr=len(keycodes_table)
            keycodes_table.append(keyplace[4])
        else:
            keynr=keyplace[4]
        kpf.write(struct.pack('<5H',keyplace[0],keyplace[1],keyplace[2],keyplace[3],keynr))
    kpf.close()
    # specification of binary format for key codes:
    # file consists of records with fixed length, each record has this format:
    # addr uint8  block len
    # addr uint8[7] keycodes for normal keypress
    # addr uint8  block len
    # addr uint8[7] keycodes for keypress with shift
    # addr uint8  block len
    # addr uint8[7] keycodes for keypress with control
    # addr uint8  block len
    # addr uint8[7] keycodes for keypress with control and shift
    def list_of_length(lst, length, pad):
        if len(lst)>=length:
            return lst[0:length]
        return lst+[pad for a in range(length-len(lst))]
    kcf=open('keyboard-'+CURRENT_KEYBOARD_SET['name']+'.keycodes','w')
    mkcb=0
    for keydef in keycodes_table:
        for kcb in keydef:
            if len(kcb)>7:
                print 'error: key code has more than 7 bytes '+str(kcb)
            kcf.write(struct.pack('<B7B', len(kcb), *list_of_length(kcb, 7, 0)))
    kcf.close()

for kbset in keyboards.KEYBOARD_SETS:
    make_one_keyboard(kbset)

